# Generated by Django 3.1.7 on 2021-10-14 09:04

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("flow", "0003_add_permission_group"),
    ]

    operations = [
        migrations.AddField(
            model_name="data",
            name="process_resources",
            field=models.JSONField(default=dict),
        ),
    ]



"""Audit manager class."""

import threading
from collections import defaultdict
from enum import IntEnum
from typing import Dict, Iterator, List, NamedTuple, Optional, Set, Tuple, Type, Union

from django.db import models
from django.http import HttpRequest, HttpResponse

GLOBAL_AUDIT_MANAGER = threading.local()

Field = str
ContentType = str
ObjectId = Union[int, str]
ObjectIds = Set[ObjectId]
Fields = Set[str]


class AccessType(IntEnum):
    """Type of the access.

    The priority is determined by numeric value of attributes: higher value
    means higher priority.
    """

    NONE = 0
    READ = 1
    UPDATE = 2
    CREATE = 3
    DELETE = 4

    def __str__(self) -> str:
        """Return the string representation of the access type object."""
        return self.name


class IterableAccessItem(NamedTuple):
    """An single item obtained by iterating over AccessLogger object."""

    content_type: ContentType
    object_id: ObjectId
    access_type: AccessType
    fields: Fields


class AccessLogger:
    """Track accesses to models.

    The basic unit is field on a model and access is tracked for every field.
    """

    def __init__(self):
        """Initialize."""
        #        self._access =
        self._access: Dict[
            ContentType, Dict[ObjectId, Dict[AccessType, Fields]]
        ] = defaultdict(lambda: defaultdict(lambda: defaultdict(set)))

    def clear(self):
        """Clear all accesses."""
        self._access.clear()

    def register_access(
        self,
        content_type: ContentType,
        object_id: ObjectId,
        access_type: AccessType,
        fields: Fields,
    ):
        """Register access to the object fields."""
        self._access[content_type][object_id][access_type].update(fields)

    def __iter__(self) -> Iterator[IterableAccessItem]:
        """Return iterator over stored access entries."""
        for content_type, model_access in self._access.items():
            for object_id, field_access_types in model_access.items():
                for access_type, fields in field_access_types.items():
                    yield IterableAccessItem(
                        content_type, object_id, access_type, fields
                    )


class AuditManager:
    """Audit manager class.

    Base object registers access to the per-request instance of audit manager
    created and stored by the custom middleware.
    """

    def __init__(self):
        """Initialize AccessLogger and custom messages list."""
        self._fields: Dict[Type[models.Model], Fields] = dict()
        self._custom_messages: List[Tuple[str, tuple, dict]] = []
        self._access_loger = AccessLogger()

    def get_content_type(self, model_class: Type[models.Model]) -> str:
        """Get the content-type from the model."""
        return model_class._meta.label

    def get_model_fields(self, model_class: Type[models.Model]) -> Fields:
        """Get all the model fields.

        Cache them to avoid excessive iteration.
        """
        if model_class not in self._fields:
            self._fields[model_class] = set(
                field.name for field in model_class._meta.fields
            )
        return self._fields[model_class]

    @staticmethod
    def log_message(message: str, *args, **kwargs):
        """Log custom message."""
        AuditManager.global_instance()._custom_messages.append((message, args, kwargs))

    @staticmethod
    def register_access(
        model_class: Type[models.Model],
        object_id: ObjectId,
        access_type: AccessType,
        fields: Optional[Fields] = None,
    ):
        """Register access to the object.

        When fields are not given the access to all fields in the model is
        logged.
        """
        manager = AuditManager.global_instance()
        fields = fields or manager.get_model_fields(model_class)
        content_type = manager.get_content_type(model_class)
        manager._access_loger.register_access(
            content_type, object_id, access_type, fields
        )

    def emit(self, request: HttpRequest, response: HttpResponse):
        """Log gathered data to the audit logger."""

        from resolwe.auditlog.logger import logger

        for access_item in self._access_loger:
            logger.log_object_access(
                request,
                response,
                access_item.content_type,
                access_item.object_id,
                access_item.access_type,
                access_item.fields,
            )

        for message, args, kwargs in self._custom_messages:
            kwargs["request"] = request
            kwargs["response"] = response
            logger.info(message, *args, **kwargs)

    def reset(self):
        """Clear all stored data."""
        self._access_loger.clear()
        self._custom_messages.clear()

    @staticmethod
    def global_instance() -> "AuditManager":
        """Return the thread-scope global instance of the audit manager."""
        if not hasattr(GLOBAL_AUDIT_MANAGER, "instance"):
            GLOBAL_AUDIT_MANAGER.instance = AuditManager()
        return GLOBAL_AUDIT_MANAGER.instance



"""Base model with methods for logging access."""
from typing import List

from django.db import models

from resolwe.auditlog.auditmanager import AccessType, AuditManager


class AuditModel(models.Model):
    """Abstract model that includes access logging methods."""

    class Meta:
        """BaseModel Meta options."""

        abstract = True

    @classmethod
    def from_db(cls, db, field_names: List[str], values: List):
        """Override from_db to register read access."""
        field_values = dict(zip(field_names, values))
        id_value = -1
        id_field_names = ["id", "uuid", "user_id"]
        for id_field_name in id_field_names:
            if id_field_name in field_values:
                id_value = field_values[id_field_name]
                break
        AuditManager.register_access(cls, id_value, AccessType.READ, set(field_names))
        return super().from_db(db, field_names, values)

    def save(self, *args, **kwargs):
        """Save the model."""
        access_type = AccessType.CREATE if self.pk is None else AccessType.UPDATE
        AuditManager.register_access(
            self._meta.model, self.pk, access_type, kwargs.get("update_fields")
        )
        super().save(*args, **kwargs)



"""Resolwe relation serializer."""
from django.db import transaction

from rest_framework import serializers

from resolwe.flow.models.collection import Collection
from resolwe.flow.models.descriptor import DescriptorSchema
from resolwe.flow.models.entity import Relation, RelationPartition, RelationType
from resolwe.flow.serializers import CollectionSerializer, DescriptorSchemaSerializer
from resolwe.flow.serializers.fields import DictRelatedField
from resolwe.permissions.models import Permission
from resolwe.rest.fields import ProjectableJSONField
from resolwe.rest.serializers import SelectiveFieldMixin

from .base import ResolweBaseSerializer


class RelationPartitionSerializer(SelectiveFieldMixin, serializers.ModelSerializer):
    """Serializer for RelationPartition objects."""

    class Meta:
        """RelationPartitionSerializer Meta options."""

        model = RelationPartition
        fields = ("id", "entity", "position", "label")


class RelationSerializer(ResolweBaseSerializer):
    """Serializer for Relation objects."""

    partitions = RelationPartitionSerializer(source="relationpartition_set", many=True)
    collection = DictRelatedField(
        queryset=Collection.objects.all(),
        serializer=CollectionSerializer,
        write_permission=Permission.EDIT,
    )
    type = serializers.SlugRelatedField(
        queryset=RelationType.objects.all(), slug_field="name"
    )
    descriptor = ProjectableJSONField(required=False)
    descriptor_schema = DictRelatedField(
        queryset=DescriptorSchema.objects.all(),
        serializer=DescriptorSchemaSerializer,
        allow_null=True,
        required=False,
    )

    class Meta:
        """RelationSerializer Meta options."""

        model = Relation
        read_only_fields = (
            "created",
            "id",
            "modified",
            "descriptor_dirty",
        )
        update_protected_fields = (
            "contributor",
            "type",
        )
        fields = (
            read_only_fields
            + update_protected_fields
            + (
                "collection",
                "category",
                "partitions",
                "unit",
                "descriptor",
                "descriptor_schema",
            )
        )

    def validate_partitions(self, partitions):
        """Raise validation error if list of partitions is empty."""
        if not partitions:
            raise serializers.ValidationError("List of partitions must not be empty.")

        return partitions

    def _create_partitions(self, instance, partitions):
        """Create partitions."""
        for partition in partitions:
            RelationPartition.objects.create(
                relation=instance,
                entity=partition["entity"],
                label=partition.get("label", None),
                position=partition.get("position", None),
            )

    def create(self, validated_data):
        """Create ``Relation`` object and add partitions of ``Entities``."""
        # `partitions` field is renamed to `relationpartition_set` based on source of nested serializer
        partitions = validated_data.pop("relationpartition_set")

        with transaction.atomic():
            instance = Relation.objects.create(**validated_data)
            self._create_partitions(instance, partitions)

        return instance

    def update(self, instance, validated_data):
        """Update ``Relation``."""
        # `partitions` field is renamed to `relationpartition_set` based on source of nested serializer
        partitions = validated_data.pop("relationpartition_set", None)

        with transaction.atomic():
            instance = super().update(instance, validated_data)

            if partitions is not None:
                # TODO: Apply the diff instead of recreating all objects.
                instance.relationpartition_set.all().delete()
                self._create_partitions(instance, partitions)

        return instance



"""Main standalone execution stub, used when the executor is run.

It should be run as a module with one argument: the relative module name
of the concrete executor class to use. The current working directory
should be where the ``executors`` module directory is, so that it can be
imported with python's ``-m <module>`` interpreter option.

Usage format:

.. code-block:: none

    /path/to/python -m executors .executor_type

Concrete example, run from the directory where ``./executors/`` is:

.. code-block:: none

    /venv/bin/python -m executors .docker

using the python from the ``venv`` virtualenv.

.. note::

    The startup code adds the concrete class name as needed, so that in
    the example above, what's actually instantiated is
    ``.docker.run.FlowExecutor``.
"""

import argparse
import asyncio
import logging
import sys
import traceback
from contextlib import suppress
from importlib import import_module

import zmq
import zmq.asyncio

from .connectors import connectors
from .global_settings import initialize_constants
from .logger import configure_logging
from .socket_utils import Response, ResponseStatus
from .zeromq_utils import ZMQCommunicator

logger = logging.getLogger(__name__)


def handle_exception(exc_type, exc_value, exc_traceback):
    """Log unhandled exceptions."""
    message = "".join(traceback.format_exception(exc_type, exc_value, exc_traceback))
    logger.error("Unhandled exception in executor: {}".format(message))


sys.excepthook = handle_exception


def validate_response(response: Response):
    """Check the response status and raise excection on error.

    :raise RuntimeError: when response has error status.
    """
    if response.response_status == ResponseStatus.ERROR:
        raise RuntimeError(
            f"Response with error status received: '{response.message_data}'."
        )


async def open_listener_connection(data_id, host, port, protocol) -> ZMQCommunicator:
    """Connect to the listener service."""
    zmq_context = zmq.asyncio.Context.instance()
    zmq_socket = zmq_context.socket(zmq.DEALER)
    zmq_socket.setsockopt(zmq.IDENTITY, f"-{data_id}".encode())
    connect_string = f"{protocol}://{host}:{port}"
    zmq_socket.connect(connect_string)
    null_logger = logging.getLogger("Docker executor<->Listener")
    null_logger.propagate = False
    null_logger.handlers = []
    return ZMQCommunicator(zmq_socket, "executor<->listener", null_logger)


async def _run_executor():
    """Start the actual execution; instantiate the executor and run."""
    parser = argparse.ArgumentParser(description="Run the specified executor.")
    parser.add_argument(
        "module", help="The module from which to instantiate the concrete executor."
    )
    parser.add_argument(
        "data_id",
        type=int,
        help="The ID of the data object this executor will process.",
    )
    parser.add_argument("host", help="The address of the listener to connect to.")
    parser.add_argument(
        "port", type=int, help="The port of the listener to connect to."
    )
    parser.add_argument("protocol", help="The protocol of the listener to connect to.")

    args = parser.parse_args()

    communicator = await open_listener_connection(
        args.data_id, args.host, args.port, args.protocol
    )
    # Start listening for responses.
    asyncio.ensure_future(communicator.start_listening())
    configure_logging(communicator)

    response = await communicator.bootstrap((args.data_id, "executor"))
    validate_response(response)
    initialize_constants(args.data_id, response.message_data)
    connectors.recreate_connectors()

    module_name = "{}.run".format(args.module)
    class_name = "FlowExecutor"

    try:
        module = import_module(module_name, __package__)
        executor = getattr(module, class_name)(
            args.data_id, communicator, (args.host, args.port, args.protocol)
        )
        await executor.run()
    except:
        logger.exception("Unexpected exception while running executor %s.", module_name)


async def _close_tasks(pending_tasks, timeout=5):
    """Close still pending tasks."""
    if pending_tasks:
        # Give tasks time to cancel.
        with suppress(asyncio.CancelledError):
            await asyncio.wait_for(asyncio.gather(*pending_tasks), timeout=timeout)
            await asyncio.gather(*pending_tasks)


async def start():
    """Start and stop all the services."""
    try:
        await _run_executor()
    finally:
        # TODO: remove asyncio.Task.all_tasks when we stop supporting Python 3.6.
        all_tasks = getattr(asyncio, "all_tasks", None) or asyncio.Task.all_tasks
        await _close_tasks(all_tasks())


if __name__ == "__main__":
    # TODO: remove this condition when we stop supporting Python 3.6.
    if sys.version_info < (3, 7):
        loop = asyncio.get_event_loop()
        loop.run_until_complete(start())
        loop.close()
    else:
        asyncio.run(start())



"""Bootstrap the containers."""
import copy
import json
import logging
import os
import shlex
import threading
from collections import defaultdict
from importlib import import_module
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Tuple, Union

from django.conf import settings
from django.core.exceptions import PermissionDenied
from django.forms.models import model_to_dict

from resolwe.flow.engine import BaseEngine, InvalidEngineError, load_engines
from resolwe.flow.executors import constants
from resolwe.flow.executors.socket_utils import Message, Response
from resolwe.flow.managers.protocol import ExecutorFiles
from resolwe.flow.models import Data, Process, Worker
from resolwe.storage import settings as storage_settings
from resolwe.test.utils import is_testing
from resolwe.utils import BraceMessage as __

from .plugin import ListenerPlugin

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from resolwe.flow.managers.listener.listener import Processor


class BootstrapCommands(ListenerPlugin):
    """Basic listener handlers."""

    def __init__(self) -> None:
        """Initialize."""
        super().__init__()
        self._executor_preparer = self._load_executor_preparer()
        self._get_program_lock = threading.Lock()
        self._bootstrap_cache: Dict[str, Any] = defaultdict(dict)
        self._execution_engines = self._load_execution_engines()
        logger.info(
            __(
                "Found {} execution engines: {}",
                len(self._execution_engines),
                ", ".join(self._execution_engines.keys()),
            )
        )
        self._expression_engines = self._load_expression_engines()
        logger.info(
            __(
                "Found {} expression engines: {}",
                len(self._expression_engines),
                ", ".join(self._expression_engines.keys()),
            )
        )

    def _bootstrap_prepare_static_cache(self):
        """Prepare cache for bootstrap."""

        def marshal_settings() -> dict:
            """Marshal Django settings into a serializable object.

            :return: The serialized settings.
            :rtype: dict
            """
            result = {}
            for key in dir(settings):
                if any(
                    map(
                        key.startswith,
                        [
                            "FLOW_",
                            "RESOLWE_",
                            "CELERY_",
                            "KUBERNETES_",
                        ],
                    )
                ):
                    result[key] = getattr(settings, key)
            result.update(
                {
                    "USE_TZ": settings.USE_TZ,
                    "FLOW_EXECUTOR_TOOLS_PATHS": self._executor_preparer.get_tools_paths(),
                    "FLOW_STORAGE": storage_settings.FLOW_STORAGE,
                }
            )
            # TODO: this is q&d solution for serializing Path objects.
            return json.loads(json.dumps(result, default=str))

        # Prepare Django settings.
        if "settings" not in self._bootstrap_cache:
            logger.debug("Preparing settings static cache.")
            self._bootstrap_cache["settings"] = marshal_settings()
            logger.debug("Settings static cache marshalled.")
            connectors_settings = copy.deepcopy(storage_settings.STORAGE_CONNECTORS)
            for connector_settings in connectors_settings.values():
                # Fix class name for inclusion in the executor.
                klass = connector_settings["connector"]
                klass = "executors." + klass.rsplit(".storage.")[-1]
                connector_settings["connector"] = klass
                connector_config = connector_settings["config"]
                # Prepare credentials for executor.
                if "credentials" in connector_config:
                    src_credentials = connector_config["credentials"]
                    base_credentials_name = os.path.basename(src_credentials)

                    self._bootstrap_cache["connector_secrets"][
                        base_credentials_name
                    ] = ""
                    if os.path.isfile(src_credentials):
                        with open(src_credentials, "r") as f:
                            self._bootstrap_cache["connector_secrets"][
                                base_credentials_name
                            ] = f.read()
                    connector_config["credentials"] = os.fspath(
                        constants.SECRETS_VOLUME / base_credentials_name
                    )
            logger.debug("Connector settings prepared.")
            self._bootstrap_cache["settings"][
                "STORAGE_CONNECTORS"
            ] = connectors_settings
            self._bootstrap_cache["settings"][
                "FLOW_VOLUMES"
            ] = storage_settings.FLOW_VOLUMES

            # Prepare process meta data.
            self._bootstrap_cache["process_meta"] = {
                k: getattr(Process, k)
                for k in dir(Process)
                if k.startswith("SCHEDULING_CLASS_")
                and isinstance(getattr(Process, k), str)
            }
            logger.debug("Process settings prepared.")
            self._bootstrap_cache["process"] = dict()

    def bootstrap_prepare_process_cache(self, data: Data):
        """Prepare cache for process with the given id."""
        if data.process_id not in self._bootstrap_cache["process"]:
            self._bootstrap_cache["process"][data.process_id] = model_to_dict(
                data.process
            )
            self._bootstrap_cache["process"][data.process_id][
                "resource_limits"
            ] = data.process.get_resource_limits()

    def handle_init_completed(
        self, data_id: int, message: Message[str], manager: "Processor"
    ) -> Response[str]:
        """Handle init completed request."""
        worker = manager.worker(data_id)
        worker.status = Worker.STATUS_FINISHED_PREPARING
        worker.save(update_fields=["status"])
        return message.respond_ok("")

    def handle_bootstrap(
        self, data_id: int, message: Message[Tuple[int, str]], manager: "Processor"
    ) -> Response[Dict]:
        """Handle bootstrap request.

        :raises RuntimeError: when settings name is not known.
        """
        data_id, settings_name = message.message_data
        logger.debug(
            __("Bootstraping peer for id {} for settings {}.", data_id, settings_name)
        )
        data = Data.objects.get(pk=data_id)
        logger.debug(__("Read data for peer with id {}.", data_id))

        if is_testing():
            self._bootstrap_cache = defaultdict(dict)
        self._bootstrap_prepare_static_cache()
        logger.debug(__("Prepared static cache for peer {}.", data_id))

        response: Dict[str, Any] = dict()
        self.bootstrap_prepare_process_cache(data)
        logger.debug(__("Prepared process cache for peer {}.", data_id))

        if settings_name == "executor":
            response[ExecutorFiles.EXECUTOR_SETTINGS] = {
                "DATA_DIR": data.location.get_path()
            }
            response[ExecutorFiles.LOCATION_SUBPATH] = data.location.subpath
            response[ExecutorFiles.DJANGO_SETTINGS] = self._bootstrap_cache[
                "settings"
            ].copy()
            response[ExecutorFiles.PROCESS_META] = self._bootstrap_cache["process_meta"]
            response[ExecutorFiles.PROCESS] = self._bootstrap_cache["process"][
                data.process.id
            ]

        elif settings_name == "init":
            response[ExecutorFiles.DJANGO_SETTINGS] = {
                "STORAGE_CONNECTORS": self._bootstrap_cache["settings"][
                    "STORAGE_CONNECTORS"
                ],
                "FLOW_STORAGE": storage_settings.FLOW_STORAGE,
                "FLOW_VOLUMES": storage_settings.FLOW_VOLUMES,
            }

            if hasattr(settings, constants.INPUTS_VOLUME_NAME):
                response[ExecutorFiles.DJANGO_SETTINGS][
                    constants.INPUTS_VOLUME_NAME
                ] = self._bootstrap_cache["settings"][constants.INPUTS_VOLUME_NAME]
            response[ExecutorFiles.SECRETS_DIR] = self._bootstrap_cache[
                "connector_secrets"
            ]
            try:
                response[ExecutorFiles.SECRETS_DIR].update(data.resolve_secrets())
            except PermissionDenied as e:
                data.process_error.append(str(e))
                data.save(update_fields=["process_error"])
                raise
            response[ExecutorFiles.LOCATION_SUBPATH] = data.location.subpath
        elif settings_name == "communication":
            response[ExecutorFiles.DJANGO_SETTINGS] = {
                "STORAGE_CONNECTORS": self._bootstrap_cache["settings"][
                    "STORAGE_CONNECTORS"
                ],
                "FLOW_STORAGE": storage_settings.FLOW_STORAGE,
            }
            response[ExecutorFiles.LOCATION_SUBPATH] = data.location.subpath
        else:
            raise RuntimeError(
                f"Settings {settings_name} sent by peer with id {data_id} unknown."
            )
        return message.respond_ok(response)

    def handle_get_script(
        self, data_id: int, message: Message[str], manager: "Processor"
    ) -> Response[str]:
        """Return script for the current Data object."""
        return message.respond_ok(self.get_program(data_id))

    def handle_log(
        self,
        data_id: int,
        message: Message[Union[str, bytes, bytearray]],
        manager: "Processor",
    ) -> Response[str]:
        """Handle an incoming log processing request.

        :param obj: The Channels message object. Command object format:

            .. code-block:: none

                {
                    'command': 'log',
                    'message': [log message]
                }
        """
        record_dict = json.loads(message.message_data)
        record_dict["args"] = tuple(record_dict["args"])

        executors_dir = Path(__file__).parents[1] / "executors"
        record_dict["pathname"] = os.fspath(executors_dir / record_dict["pathname"])
        logger.handle(logging.makeLogRecord(record_dict))
        return message.respond_ok("OK")

    def get_program(self, data_id: int) -> str:
        """Get a program for given data object."""
        # When multiple get_script commands run in parallel the string
        # escaping will be fragile at best. Process this command here
        # without offloading it to the listener.
        data = Data.objects.get(pk=data_id)

        with self._get_program_lock:
            execution_engine_name = data.process.run.get("language", None)
            program = [self._get_execution_engine(execution_engine_name).evaluate(data)]
            # TODO: should executor be changed? Definitely for tests (from case to case).
            # Should I use file prepared by the Dispatcher (that takes care of that)?
            env_vars = self.get_executor().get_environment_variables()
            settings_env_vars = {
                "RESOLWE_HOST_URL": getattr(settings, "RESOLWE_HOST_URL", "localhost"),
            }
            additional_env_vars = getattr(settings, "FLOW_EXECUTOR", {}).get(
                "SET_ENV", {}
            )
            env_vars.update(settings_env_vars)
            env_vars.update(additional_env_vars)
            export_commands = [
                "export {}={}".format(key, shlex.quote(value))
                for key, value in env_vars.items()
            ]

            # Add tools to the path. The tools paths environment variable must
            # be prepared by the executor.
            export_commands += ["export PATH=$PATH:$TOOLS_PATHS"]

            # Disable brace expansion and set echo.
            echo_commands = ["set -x +B"]

            commands = echo_commands + export_commands + program
            return os.linesep.join(commands)

    def _get_execution_engine(self, name):
        """Return an execution engine instance."""
        try:
            return self._execution_engines[name]
        except KeyError:
            raise InvalidEngineError("Unsupported execution engine: {}".format(name))

    def _load_execution_engines(self) -> Dict[str, BaseEngine]:
        """Load execution engines."""
        execution_engines = getattr(
            settings, "FLOW_EXECUTION_ENGINES", ["resolwe.flow.execution_engines.bash"]
        )
        return load_engines(
            self, "ExecutionEngine", "execution_engines", execution_engines
        )

    def _load_expression_engines(self) -> Dict[str, BaseEngine]:
        """Load expression engines."""
        expression_engines = getattr(
            settings,
            "FLOW_EXPRESSION_ENGINES",
            ["resolwe.flow.expression_engines.jinja"],
        )
        return load_engines(
            self, "ExpressionEngine", "expression_engines", expression_engines
        )

    def _load_executor_preparer(self) -> Any:
        """Load and return the executor preparer class."""
        executor_name = (
            getattr(settings, "FLOW_EXECUTOR", {}).get(
                "NAME", "resolwe.flow.executors.docker"
            )
            + ".prepare"
        )
        return import_module(executor_name).FlowExecutorPreparer()

    def get_executor(self) -> Any:
        """Get the executor preparer class."""
        return self._executor_preparer

    def get_expression_engine(self, name: str):
        """Return an expression engine instance."""
        try:
            return self._expression_engines[name]
        except KeyError:
            raise InvalidEngineError("Unsupported expression engine: {}".format(name))



""".. Ignore pydocstyle D400.

==================
Resolwe Decorators
==================

Utils functions.

"""

import logging
import time
from typing import Callable, Optional, Sequence, Type

import wrapt


def retry(
    logger: Optional[logging.Logger],
    max_retries: int = 3,
    retry_exceptions: Sequence[Type[Exception]] = (Exception,),
    min_sleep: int = 1,
    max_sleep: int = 10,
    cleanup_callback: Optional[Callable] = None,
):
    """Retry decorator.

    A method decorated with it will be called up to ``max_retries`` times with
    an exponential timeout ranging from ``min_sleep`` to ``max_sleep`` between
    retries. The method will be retried if it raises one of the exception in
    the ``retry_exceptions`` sequence.

    :attr cleanup_callback: the callback to call after exception and the next
        retry of the decorated function. It is given the same arguments as the
        decorated function. Any exception raised by the cleanup_handler is
        logged but not raised.

    :raises Exception: when all retries raise exception the last one is
        re-raised.
    """

    @wrapt.decorator
    def wrapper(wrapped, instance, args, kwargs):
        """Retry on tranfser error."""
        for retry in range(1, max_retries + 1):
            try:
                return wrapped(*args, **kwargs)
            except retry_exceptions as exception:
                # Log the exception on retry for inspection.
                if retry < max_retries:
                    sleep = min(max_sleep, min_sleep * (2 ** (retry - 1)))
                    if logger is not None:
                        logger.exception(
                            "Retry %d/%d got exception, will retry in %.2f seconds.",
                            retry,
                            max_retries,
                            sleep,
                        )
                    if cleanup_callback is not None:
                        try:
                            cleanup_callback(*args, **kwargs)
                        except Exception:
                            logger.exception("Exception in callback handler.")
                    time.sleep(sleep)
                # Raise exception if all retries have failed.
                else:
                    if logger is not None:
                        logger.exception(
                            "Retry %d/%d got exception, re-raising it.",
                            retry,
                            max_retries,
                        )
                    raise exception

    return wrapper



"""Entity viewset."""
from django.core.exceptions import ValidationError as DjangoValidationError
from django.db import transaction
from django.db.models import F, Func, OuterRef, Prefetch, Subquery
from django.db.models.functions import Coalesce

from rest_framework import exceptions
from rest_framework.decorators import action
from rest_framework.response import Response

from resolwe.flow.filters import EntityFilter
from resolwe.flow.models import AnnotationValue, Data, DescriptorSchema, Entity
from resolwe.flow.models.annotations import AnnotationField, HandleMissingAnnotations
from resolwe.flow.serializers import EntitySerializer
from resolwe.flow.serializers.annotations import (
    AnnotationsSerializer,
    AnnotationValueSerializer,
)
from resolwe.observers.mixins import ObservableMixin
from resolwe.observers.views import BackgroundTaskSerializer
from resolwe.process.descriptor import ValidationError

from .collection import BaseCollectionViewSet
from .utils import get_collection_for_user


class EntityViewSet(ObservableMixin, BaseCollectionViewSet):
    """API view for entities."""

    serializer_class = EntitySerializer
    filter_class = EntityFilter
    qs_descriptor_schema = DescriptorSchema.objects.select_related("contributor")

    data_count_subquery = (
        Data.objects.filter(entity_id=OuterRef("pk"))
        .annotate(count=Func(F("id"), function="Count"))
        .values("count")
    )
    data_status_subquery = (
        Data.objects.filter(entity_id=OuterRef("pk"))
        .annotate(
            count=Coalesce(
                Func(
                    F("status"),
                    function="ARRAY_AGG",
                    template="%(function)s(DISTINCT %(expressions)s)",
                ),
                [],
            )
        )
        .values("count")
    )
    queryset = (
        Entity.objects.select_related("contributor", "descriptor_schema__contributor")
        .prefetch_related(
            Prefetch("collection", queryset=BaseCollectionViewSet.queryset)
        )
        .annotate(data_statuses=Subquery(data_status_subquery))
        .annotate(data_count=Subquery(data_count_subquery))
    )

    def _get_entities(self, user, ids):
        """Return entities queryset based on provided entity ids."""
        queryset = Entity.objects.filter(id__in=ids).filter_for_user(user)
        actual_ids = queryset.values_list("id", flat=True)
        missing_ids = list(set(ids) - set(actual_ids))
        if missing_ids:
            raise exceptions.ParseError(
                "Entities with the following ids not found: {}".format(
                    ", ".join(map(str, missing_ids))
                )
            )

        return queryset

    @action(detail=False, methods=["post"])
    def move_to_collection(self, request, *args, **kwargs):
        """Move samples from source to destination collection.

        The request accepts argument 'handle_missing_annotations'.The possible
        values for it are 'ADD' and 'REMOVE'. When 'REMOVE' is set without
        arguments the 'confirm_annotations_delete' argument must be set to
        True.

        :raises ValidationError: when handle_missing_annotations has incorrect
            value.
        :raises ValidationError: when handle_missing_annotations is set to
            'REMOVE' and 'confirm_annotations_delete' is set.
        """

        ids = self.get_ids(request.data)
        dst_collection_id = self.get_id(request.data, "destination_collection")
        dst_collection = get_collection_for_user(dst_collection_id, request.user)
        entity_qs = self._get_entities(request.user, ids)

        missing_annotations = {
            entity: AnnotationField.objects.filter(collection=entity.collection)
            .exclude(collection=dst_collection)
            .distinct()
            for entity in entity_qs
        }

        if missing_annotations:
            # How to handle annotation fields present in the current collection
            # but missing in new one. We have two options:
            # - add (default): add the missing annotation fields to the new
            #   collection and keep the values.
            # - delete: remove the missing annotation values from the samples.
            #   This option requires explicit confirmation by 'confirm_remove'
            #   argument or exception with explanation is raised.
            try:
                handle_missing_annotations = HandleMissingAnnotations(
                    request.data.get(
                        "handle_missing_annotations", HandleMissingAnnotations.ADD.name
                    )
                )
            except ValueError:
                possible_values = HandleMissingAnnotations._member_names_
                raise ValidationError(
                    "Attribute 'handle_missing_annotations' must be one of "
                    f"{possible_values}."
                )

            if handle_missing_annotations == HandleMissingAnnotations.REMOVE:
                if request.data.get("confirm_annotations_remove", False) is not True:
                    raise ValidationError(
                        "All annotations not present in the target collection will be "
                        "removed. Set 'confirm_action' argument to 'True' to confirm."
                    )

        entity_qs.move_to_collection(dst_collection, handle_missing_annotations)
        return Response()

    # NOTE: This can be deleted when DRF will support select_for_update
    #       on updates and ResolweUpdateModelMixin will use it.
    #       https://github.com/encode/django-rest-framework/issues/4675
    def update(self, request, *args, **kwargs):
        """Update an entity.

        Original queryset produces a temporary database table whose rows
        cannot be selected for an update. As a workaround, we patch
        get_queryset function to return only Entity objects without
        additional data that is not needed for the update.
        """
        orig_get_queryset = self.get_queryset

        def patched_get_queryset():
            """Patched get_queryset method."""
            entity_ids = orig_get_queryset().values_list("id", flat=True)
            return Entity.objects.filter(id__in=entity_ids)

        self.get_queryset = patched_get_queryset

        resp = super().update(request, *args, **kwargs)

        self.get_queryset = orig_get_queryset
        return resp

    @action(detail=False, methods=["post"])
    def duplicate(self, request, *args, **kwargs):
        """Duplicate (make copy of) ``Entity`` models.

        The objects are duplicated in the background and the details of the background
        task handling the duplication are returned.
        """
        if not request.user.is_authenticated:
            raise exceptions.NotFound

        inherit_collection = request.data.get("inherit_collection", False)
        ids = self.get_ids(request.data)
        queryset = Entity.objects.filter(id__in=ids).filter_for_user(request.user)
        actual_ids = queryset.values_list("id", flat=True)
        missing_ids = list(set(ids) - set(actual_ids))
        if missing_ids:
            raise exceptions.ParseError(
                "Entities with the following ids not found: {}".format(
                    ", ".join(map(str, missing_ids))
                )
            )

        task = queryset.duplicate(
            contributor=request.user, inherit_collection=inherit_collection
        )
        serializer = BackgroundTaskSerializer(task)
        return Response(serializer.data)

    @action(detail=True, methods=["post"])
    def set_annotations(self, request, pk=None):
        """Add the given list of AnnotaitonFields to the given collection."""
        # No need to check for permissions, since post requires edit by default.
        entity = self.get_object()
        # Read and validate the request data.
        serializer = AnnotationsSerializer(data=request.data, many=True)
        serializer.is_valid(raise_exception=True)
        annotations = [
            (entry["field"], entry["value"]) for entry in serializer.validated_data
        ]
        annotation_fields = [e[0] for e in annotations]
        # The following dict is a mapping from annotation field id to the annotation
        # value id.
        existing_annotations = dict(
            entity.annotations.filter(field__in=annotation_fields).values_list(
                "field_id", "id"
            )
        )

        validation_errors = []
        to_create = []
        to_update = []
        for field, value in annotations:
            annotation_id = existing_annotations.get(field.id)
            append_to = to_create if annotation_id is None else to_update
            annotation = AnnotationValue(
                entity_id=entity.id, field_id=field.id, value=value, id=annotation_id
            )
            try:
                annotation.validate()
            except DjangoValidationError as e:
                validation_errors += e
            append_to.append(annotation)

        if validation_errors:
            raise DjangoValidationError(validation_errors)

        with transaction.atomic():
            AnnotationValue.objects.bulk_create(to_create)
            AnnotationValue.objects.bulk_update(to_update, ["_value"])
        return Response()

    @action(detail=True, methods=["get"])
    def get_annotations(self):
        """Get the annotations on this sample."""
        entity = self.get_object()
        serializer = AnnotationValueSerializer(entity.annotations.all(), many=True)
        return Response(serializer.data)



""".. Ignore pydocstyle D400.

============
Flow Filters
============

"""
from functools import partial

from django_filters import rest_framework as filters
from django_filters.filterset import FilterSetMetaclass
from versionfield import VersionField

from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group
from django.contrib.postgres.search import SearchQuery, SearchRank
from django.db.models import Count, F, ForeignKey, Q, Subquery

from rest_framework import exceptions
from rest_framework.filters import OrderingFilter as DrfOrderingFilter

from resolwe.composer import composer
from resolwe.permissions.models import Permission, get_anonymous_user

from .models import (
    AnnotationField,
    AnnotationPreset,
    AnnotationValue,
    Collection,
    Data,
    DescriptorSchema,
    Entity,
    Process,
    Relation,
)

RELATED_LOOKUPS = [
    "exact",
    "in",
    "isnull",
]
NUMBER_LOOKUPS = [
    "exact",
    "in",
    "gt",
    "gte",
    "lt",
    "lte",
    "isnull",
]
TEXT_LOOKUPS = [
    "exact",
    "iexact",
    "contains",
    "icontains",
    "in",
    "startswith",
    "istartswith",
    "isnull",
]
DATE_LOOKUPS = [
    "exact",
    "gt",
    "gte",
    "lt",
    "lte",
    "isnull",
]
DATETIME_LOOKUPS = DATE_LOOKUPS + [
    "date",
    "time",
]
SLUG_LOOKUPS = [
    "exact",
    "in",
]


user_model = get_user_model()


class CheckQueryParamsMixin:
    """Custom query params validation."""

    def get_always_allowed_arguments(self):
        """Get always allowed query arguments."""
        return (
            "fields",
            "format",
            "limit",
            "offset",
            "ordering",
        )

    def validate_query_params(self):
        """Ensure no unsupported query params were used."""
        allowed_params = set(self.get_filters().keys())
        allowed_params.update(self.get_always_allowed_arguments())

        unallowed = set(self.request.query_params.keys()) - allowed_params

        if unallowed:
            msg = (
                "Unsupported parameter(s): {}. Please use a combination of: {}.".format(
                    ", ".join(unallowed), ", ".join(allowed_params)
                )
            )
            self.form.add_error(field=None, error=msg)

    def is_valid(self):
        """Validate filterset."""
        self.validate_query_params()
        return super().is_valid()


class TextFilterMixin:
    """Mixin for full-text filtering."""

    def filter_text(self, queryset, name, value):
        """Full-text search."""
        query = SearchQuery(value, config="simple")
        return (
            queryset.filter(**{name: query})
            # This assumes that field is already a TextSearch vector and thus
            # doesn't need to be transformed. To achieve that F function is
            # required.
            .annotate(rank=SearchRank(F(name), query)).order_by("-rank")
        )


class UserFilterMixin:
    """Mixin for filtering by contributors and owners."""

    @staticmethod
    def _get_user_subquery(value):
        user_subquery = user_model.objects.all()
        for key in value.split():
            user_subquery = user_subquery.extra(
                where=["CONCAT_WS(' ', first_name, last_name, username) ILIKE %s"],
                params=("%{}%".format(key),),
            )

        return user_subquery

    def filter_owners(self, queryset, name, value):
        """Filter queryset by owner's id."""
        try:
            user = user_model.objects.get(pk=value)
        except user_model.DoesNotExist:
            return queryset.none()

        return queryset.filter_for_user(
            user, self.owner_permission, with_superuser=False
        )

    def filter_owners_name(self, queryset, name, value):
        """Filter queryset by owner's name."""
        result = queryset.model.objects.none()
        user_subquery = self._get_user_subquery(value)
        for user in user_subquery:
            result = result.union(
                queryset.filter_for_user(
                    user, self.owner_permission, with_superuser=False
                )
            )

        # Union can no longer be filtered, so we have to create a new queryset
        # for following filters.
        return result.model.objects.filter(pk__in=Subquery(result.values("pk")))

    def filter_contributor_name(self, queryset, name, value):
        """Filter queryset by owner's name."""
        return queryset.filter(contributor__in=self._get_user_subquery(value))

    def filter_for_user(self, queryset, name, value):
        """Filter queryset by permissions."""
        user = self.request.user
        return queryset.filter_for_user(user, Permission.from_name(value))

    def filter_for_group(self, queryset, name, value):
        """Filter queryset by group."""
        return queryset.filter(permission_group__permissions__group=value)

    def filter_private(self, queryset, name, value):
        """Return only elements that are not public (explicitly shared with me)."""

        public_filter = Q(permission_group__permissions__user=get_anonymous_user())

        # Exclude public elements if filter value is True
        if value:
            return queryset.exclude(public_filter)
        else:
            return queryset.filter(public_filter)


class TagsFilter(filters.BaseCSVFilter, filters.CharFilter):
    """Filter for tags."""

    def __init__(self, *args, **kwargs):
        """Construct tags filter."""
        kwargs.setdefault("lookup_expr", "contains")
        super().__init__(*args, **kwargs)


class ResolweFilterMetaclass(FilterSetMetaclass):
    """Support extending filter by third-party packages.

    .. note::

        This injects filters in an undocumented way and can thus break in
        future releases.

    For example, to add a filter by ``my_field``, define an extension and add
    it to the composer:

    .. code-block:: python

        from resolwe.composer import composer


        class ExtendedFilter:

            my_field = filters.CharFilter()

        composer.add_extension("path.to.my.Filter", ExtendedFilter)

    """

    def __new__(mcs, name, bases, namespace):
        """Inject extensions into the filter."""
        class_path = "{}.{}".format(namespace["__module__"], name)
        for extension in composer.get_extensions(class_path):
            for name in dir(extension):
                if name.startswith("__"):
                    continue
                namespace[name] = getattr(extension, name)

        return super().__new__(mcs, name, bases, namespace)


class BaseResolweFilter(
    CheckQueryParamsMixin, filters.FilterSet, metaclass=ResolweFilterMetaclass
):
    """Base filter for Resolwe's endpoints."""

    owner_permission = Permission.OWNER

    class Meta:
        """Filter configuration."""

        fields = {
            "contributor": ["exact", "in"],
            "created": DATETIME_LOOKUPS[:],
            "id": NUMBER_LOOKUPS[:],
            "name": TEXT_LOOKUPS[:],
            "modified": DATETIME_LOOKUPS[:],
            "slug": SLUG_LOOKUPS[:],
            "version": NUMBER_LOOKUPS[:],
        }

        filter_overrides = {
            VersionField: {"filter_class": filters.CharFilter},
            ForeignKey: {"filter_class": filters.Filter},
        }


class DescriptorSchemaFilter(BaseResolweFilter):
    """Filter the DescriptorSchema endpoint."""

    class Meta(BaseResolweFilter.Meta):
        """Filter configuration."""

        model = DescriptorSchema


class BaseCollectionFilter(TextFilterMixin, UserFilterMixin, BaseResolweFilter):
    """Base filter for Collection and Entity endpoints."""

    contributor_name = filters.CharFilter(method="filter_contributor_name")
    owners = filters.CharFilter(method="filter_owners")
    owners_name = filters.CharFilter(method="filter_owners_name")
    shared_with_me = filters.BooleanFilter(method="filter_private")
    group = filters.ModelChoiceFilter(
        method="filter_for_group", queryset=Group.objects.all()
    )
    permission = filters.CharFilter(method="filter_for_user")
    tags = TagsFilter()
    text = filters.CharFilter(field_name="search", method="filter_text")

    class Meta(BaseResolweFilter.Meta):
        """Filter configuration."""

        fields = {
            **BaseResolweFilter.Meta.fields,
            **{
                "description": TEXT_LOOKUPS[:],
                "descriptor_schema": ["exact"],
            },
        }


class CollectionFilter(BaseCollectionFilter):
    """Filter the Collection endpoint."""

    entity_count = filters.NumberFilter(
        method="count_entities", field_name="entity__count"
    )
    entity_count__gt = filters.NumberFilter(
        method="count_entities", field_name="entity__count__gt"
    )
    entity_count__lt = filters.NumberFilter(
        method="count_entities", field_name="entity__count__lt"
    )
    entity_count__gte = filters.NumberFilter(
        method="count_entities", field_name="entity__count__gte"
    )
    entity_count__lte = filters.NumberFilter(
        method="count_entities", field_name="entity__count__lte"
    )

    class Meta(BaseCollectionFilter.Meta):
        """Filter configuration."""

        model = Collection

    def count_entities(self, queryset, name, value):
        """Filter by the number of associated entities."""

        return queryset.annotate(Count("entity")).filter(**{name: value})


class EntityFilter(BaseCollectionFilter):
    """Filter the Entity endpoint."""

    relation_id = filters.NumberFilter(field_name="relation__id")

    class Meta(BaseCollectionFilter.Meta):
        """Filter configuration."""

        model = Entity
        fields = {
            **BaseCollectionFilter.Meta.fields,
            **{
                "collection": RELATED_LOOKUPS[:],
                "collection__name": TEXT_LOOKUPS[:],
                "collection__slug": SLUG_LOOKUPS[:],
                "type": ["exact"],
            },
        }


class ProcessFilter(BaseResolweFilter):
    """Filter the Process endpoint."""

    category = filters.CharFilter(field_name="category", lookup_expr="startswith")
    type = filters.CharFilter(field_name="type", lookup_expr="startswith")

    class Meta(BaseResolweFilter.Meta):
        """Filter configuration."""

        model = Process
        fields = {
            **BaseResolweFilter.Meta.fields,
            **{
                "is_active": ["exact"],
                "scheduling_class": ["exact"],
            },
        }


class CharInFilter(filters.BaseInFilter, filters.CharFilter):
    """Helper class for creation of CharFilter with "in" lookup."""


class DataFilter(TextFilterMixin, UserFilterMixin, BaseResolweFilter):
    """Filter the Data endpoint."""

    contributor_name = filters.CharFilter(method="filter_contributor_name")
    owners = filters.CharFilter(method="filter_owners")
    owners_name = filters.CharFilter(method="filter_owners_name")
    tags = TagsFilter()
    text = filters.CharFilter(field_name="search", method="filter_text")
    type = filters.CharFilter(field_name="process__type", lookup_expr="startswith")
    type__exact = filters.CharFilter(field_name="process__type", lookup_expr="exact")
    relation_id = filters.NumberFilter(
        field_name="entity__relationpartition__relation_id"
    )
    shared_with_me = filters.BooleanFilter(method="filter_private")
    group = filters.ModelChoiceFilter(
        method="filter_for_group", queryset=Group.objects.all()
    )
    permission = filters.CharFilter(method="filter_for_user")

    class Meta(BaseResolweFilter.Meta):
        """Filter configuration."""

        model = Data
        fields = {
            **BaseResolweFilter.Meta.fields,
            **{
                "collection": RELATED_LOOKUPS[:],
                "collection__name": TEXT_LOOKUPS[:],
                "collection__slug": SLUG_LOOKUPS[:],
                "entity": RELATED_LOOKUPS[:],
                "entity__name": TEXT_LOOKUPS[:],
                "entity__slug": SLUG_LOOKUPS[:],
                "process": RELATED_LOOKUPS[:],
                "process__name": TEXT_LOOKUPS[:],
                "process__slug": SLUG_LOOKUPS[:],
                "finished": DATETIME_LOOKUPS[:],
                "started": DATETIME_LOOKUPS[:],
                "status": ["exact", "in"],
            },
        }


class RelationFilter(BaseResolweFilter):
    """Filter the Relation endpoint."""

    category = filters.CharFilter(lookup_expr="iexact")
    collection = filters.ModelChoiceFilter(queryset=Collection.objects.all())
    type = filters.CharFilter(field_name="type__name")
    tags = TagsFilter(field_name="collection__tags")

    class Meta(BaseResolweFilter.Meta):
        """Filter configuration."""

        model = Relation
        fields = {
            **BaseResolweFilter.Meta.fields,
            **{
                "descriptor_schema": ["exact"],
            },
        }

    def get_always_allowed_arguments(self):
        """Get always allowed query arguments."""
        return super().get_always_allowed_arguments() + (
            "entity",
            "label",
            "position",
        )


class OrderingFilter(DrfOrderingFilter):
    """Order results by field specified in request.

    If no ordering is given, default ordering is used. Default ordering
    is skipped for views filtered by full-text search as it already
    applies ordering by ranking.
    """

    def get_ordering(self, request, queryset, view):
        """Return name of the filed by which results should be ordered.

        Ordering is skipped if ``None`` is returned.
        """
        params = request.query_params.get(self.ordering_param)
        if params:
            fields = [param.strip() for param in params.split(",")]
            ordering = self.remove_invalid_fields(queryset, fields, view, request)
            if ordering:
                return ordering

        # Skip default ordering as full-text search applies ordering by rank.
        if request.query_params.get("text"):
            return None

        return self.get_default_ordering(view)


class AnnotationFieldFilter(BaseResolweFilter):
    """Filter the AnnotationField endpoint."""

    @classmethod
    def filter_for_field(cls, field, field_name, lookup_expr=None):
        """Add permission check for collections lookups.

        The old filter method is moved to the original_filter and replaced with the
        filter that checks the permissions on the collections.
        """

        def collection_permission_filter(self, qs, value):
            """Check if user has access the collections.

            The collection filer is defined in the parent filter.

            :raises PermissionDenied: if user does not have the permissions to access any
                of the collections defined by the filter.
            """

            if value:
                field_name, lookup_expression = self.field_name, self.lookup_expr
                if field_name == "collection":
                    field_name += "__id"

                collection_field = field_name.split("__", maxsplit=1)[1]
                collection_filter = f"{collection_field}__{lookup_expression}"
                filtered_collections = Collection.objects.filter(
                    **{collection_filter: value}
                )
                user_visible_collections = filtered_collections.filter_for_user(
                    self.parent.request.user
                )
                if filtered_collections.count() != user_visible_collections.count():
                    raise exceptions.PermissionDenied()
            return self.original_filter(qs, value)

        base_filter = filters.FilterSet.filter_for_field(field, field_name, lookup_expr)
        # No need to overwrite the exact filter.
        if field_name.startswith("collection__"):
            base_filter.original_filter = base_filter.filter
            base_filter.filter = partial(collection_permission_filter, base_filter)
        return base_filter

    class Meta(BaseResolweFilter.Meta):
        """Filter configuration."""

        model = AnnotationField
        fields = {
            **{
                "name": TEXT_LOOKUPS,
                "label": TEXT_LOOKUPS,
                "type": TEXT_LOOKUPS,
                "description": TEXT_LOOKUPS,
                "group__name": TEXT_LOOKUPS,
                "collection": ["exact"],
                "collection__name": TEXT_LOOKUPS[:],
                "collection__slug": SLUG_LOOKUPS[:],
            },
        }


class AnnotationPresetFilter(BaseResolweFilter):
    """Filter the AnnotationPreset endpoint."""

    class Meta(BaseResolweFilter.Meta):
        """Filter configuration."""

        model = AnnotationPreset
        fields = {
            **{
                "name": TEXT_LOOKUPS,
            },
        }


class AnnotationValueFilter(BaseResolweFilter):
    """Filter the AnnotationValue endpoint."""

    entity_ids = filters.CharFilter(method="filter_by_entities", required=True)
    field_ids = filters.CharFilter(method="filter_by_fields")
    value = filters.CharFilter(method="filter_by_value")

    def filter_by_entities(self, queryset, name, entity_ids):
        """Filter by the collection id.

        The filtering must also take permissions on the collections in account.
        """
        entities = Entity.objects.filter(pk__in=entity_ids).filter_for_user(
            self.request.user
        )
        return queryset.filter(entity__in=entities)

    def filter_by_fields(self, queryset, name, field_ids):
        """Filter by field ids."""
        return queryset.filter(field_id__in=field_ids)

    def filter_by_value(self, queryset, name, value):
        """Filter by value."""
        return queryset.filter(_value__value__icontains=value)

    class Meta(BaseResolweFilter.Meta):
        """Filter configuration."""

        model = AnnotationValue
        fields = {
            **{"field__name": TEXT_LOOKUPS},
            **{"field__label": TEXT_LOOKUPS},
        }



""".. Ignore pydocstyle D400.

===============
Signal Handlers
===============

"""
from typing import List, Optional

from asgiref.sync import async_to_sync

from django.conf import settings
from django.db import transaction
from django.db.models.signals import post_delete, post_save
from django.dispatch import Signal, receiver

from resolwe.flow.managers import manager
from resolwe.flow.models import Data, Relation
from resolwe.flow.models.entity import RelationPartition

# The signal sent before processing of the Data object starts. It is useful to
# attach to it if some other models etc have to be created before the actual
# processing takes place.
before_processing = Signal()

# Sent when data objects are copied in bulk (for instance when duplicating
# data objects) and no other signals are sent.
post_duplicate = Signal()


def commit_signal(data: Data, created: bool, update_fields: Optional[List[str]]):
    """Nudge manager at the end of every Data object save event.

    Also sent before_processing signal just before nudging manager.
    """
    if not getattr(settings, "FLOW_MANAGER_DISABLE_AUTO_CALLS", False):
        before_processing.send(
            sender=Data, instance=data, created=created, update_fields=update_fields
        )
        immediate = getattr(settings, "FLOW_MANAGER_SYNC_AUTO_CALLS", False)
        async_to_sync(manager.communicate)(data_id=data.id, run_sync=immediate)


@receiver(post_save, sender=Data)
def manager_post_save_handler(sender, instance: Data, created: bool, **kwargs):
    """Run newly created (spawned) processes."""
    if (
        instance.status == Data.STATUS_DONE
        or instance.status == Data.STATUS_ERROR
        or created
    ):
        # Run manager at the end of the potential transaction. Otherwise
        # tasks are send to workers before transaction ends and therefore
        # workers cannot access objects created inside transaction.
        transaction.on_commit(
            lambda: commit_signal(instance, created, kwargs.get("update_fields"))
        )


# NOTE: m2m_changed signal cannot be used because of a bug:
# https://code.djangoproject.com/ticket/17688
@receiver(post_delete, sender=RelationPartition)
def delete_relation(sender, instance: RelationPartition, **kwargs):
    """Delete the Relation object when the last Entity is removed."""

    def process_signal(relation_id):
        """Get the relation and delete it if it has no entities left."""
        try:
            relation = Relation.objects.get(pk=relation_id)
        except Relation.DoesNotExist:
            return

        if relation.entities.count() == 0:
            relation.delete()

    # Wait for partitions to be recreated.
    transaction.on_commit(lambda: process_signal(instance.relation_id))
